/* This is the template file for HW2_7. You need to complete the code by 
 * implementing the parts marked by FIXME. 
 *
 * Submit your completed code to the TA by email. Also submit the plots 
 * and your conclusions of the comparative study.
 *
 * To compile, use the UNIX command 
 *      make
 * in the source directory, and it would then invoke the compilation command 
 * with the included makefile. You may sometimes need to use the command 
 *      make clean
 * before recompiling.
 *
 * This will generate an executable hw2_7. The -g option is optional and 
 * is needed only if you need to debug the program in a debugger such as ddd.
 * The -Wall option would enable compiler's warning messages.
 *
 * Run the program with command
 *      ./hw2_7
 * which would generate a M-file, which you can use to generate the plots by
 * issueing the UNIX command 
 *      make plot
 * in the source directory. It will then generate two PDF files, which you
 * should submit to the TA along with your source code and your conclusions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

/* Define types for matrix and vector. 
 * Access the (i,j)-entries of a matrix A using A.vals[i][j], and 
 * access the ith entries of a vector x using x.vals[i], where 
 * i and j start from 0.
 */

typedef struct {
    double** vals;
    int m; /* Number of rows */
    int n; /* Number of columns */
} Matrix;

typedef struct {
    double*  vals;
    int m;
} Vector;

/* Prototypes for local functions. 
   You need to implement some of these functions. */
void  lu_no_pivot(Matrix A, Vector x, const Vector b);
void  lu_partial_pivot(Matrix A, Vector x, const Vector b);
void  lu_complete_pivot(Matrix A, Vector x, const Vector b, Vector work);

void  forward_solve( Vector x, const Matrix L, const Vector b);
void  back_solve( Vector x, const Matrix U, const Vector b);

void   permute_vec( Vector y, int *P, const Vector x, int inv);
double norm2_col( double *vec, int vec_dim);

void rand_matrix_vector( Matrix A, Vector b, int seed);
void compute_error( const Matrix A, const Vector x, const Vector b, 
                    double *err, double *res);

/* Define prototypes of helper functions */
Matrix allocate_matrix(int m, int n);
Vector allocate_vector( int m);
void   deallocate_matrix( Matrix);
void   deallocate_vector( Vector);

void   write_vec(FILE *fid, char *name, const double *vec, int n);
double tictoc (int n);

#define MAXDIM    200

int main(int argc, char **argv) 
{   
    double res_np[MAXDIM], res_pp[MAXDIM], res_cp[MAXDIM];
    double err_np[MAXDIM], err_pp[MAXDIM], err_cp[MAXDIM];
    double times_np[MAXDIM], times_pp[MAXDIM], times_cp[MAXDIM];

    FILE *fid;
    int m, n;

    /* Set n to rank of matrix */
    for (n=3; n<=MAXDIM; ++n) {
        Matrix A; 
        Vector b,x,y;

        /* Set m to the number of points */
        m = n;

        A = allocate_matrix( m, m);
        b = allocate_vector( m);
        x = allocate_vector( m);
        y = allocate_vector( m);

        /* Generate random matrix and right-hand-side vector with differnt seeds. */
        rand_matrix_vector(A, b, n);

        /* Invoke lu_no_pivot.*/
        tictoc(1);
        lu_no_pivot( A, x, b);
        times_np[n - 3] =  tictoc(2);

        /* Compute error and residual. Note that A has to be regenerated. */
        rand_matrix_vector(A, b, n);
        compute_error( A, x, b, &err_np[n - 3], &res_np[n - 3]);

        /* Invoke LU with partial pivoting. */
        tictoc(1);
        lu_partial_pivot( A, x, b);
        times_pp[n - 3] =  tictoc(2);

        /* Compute error and residual. Note that A has to be regenerated. */
        rand_matrix_vector(A, b, n);
        compute_error( A, x, b, &err_pp[n - 3], &res_pp[n - 3]);

        /* Invoke LU with complete pivoting. */
        tictoc(1);
        lu_complete_pivot( A, x, b, y);
        times_cp[n - 3] =  tictoc(2);

        /* Compute error and residual. Note that A has to be regenerated. */
        rand_matrix_vector(A, b, n);
        compute_error( A, x, b, &err_cp[n - 3], &res_cp[n - 3]);

        deallocate_matrix( A);
        deallocate_vector( b);
        deallocate_vector( x);
        deallocate_vector( y);
    }

    /* Write out results into a Matlab file */
    fid = fopen("results.m", "w");
    write_vec(fid, "err_np", err_np, MAXDIM-2);
    write_vec(fid, "err_pp  ", err_pp, MAXDIM-2);
    write_vec(fid, "err_cp  ", err_cp, MAXDIM-2);

    write_vec(fid, "res_np", res_np, MAXDIM-2);
    write_vec(fid, "res_pp", res_pp, MAXDIM-2);
    write_vec(fid, "res_cp", res_cp, MAXDIM-2);

    write_vec(fid, "times_np", times_np, MAXDIM-2);
    write_vec(fid, "times_pp", times_pp, MAXDIM-2);
    write_vec(fid, "times_cp", times_cp, MAXDIM-2);
    fclose(fid);

    return 0;
}

/*************************************************************
 *
 * FUNCTION: lu_no_pivot
 *
 * LU factorization without pivoting.
 * Note that A is overwritten by L and U matrices.
 *************************************************************/
void  lu_no_pivot(Matrix A, Vector x, const Vector b) {
    int i, j, k, m=A.m;

    /* First, perform LU factorization. */
    for ( k=0; k<m-1; ++k) {

        if (A.vals[k][k]==0) {
            fprintf(stderr, "Warning: LU without pivoting broke down.\n");
            continue;
        }

        /* FIXME. Refer to lu_partial_pivot */
        /* A(k+1:m,k) = A(k+1:m,k) / A(k,k); */

        /* A(k+1:m,k+1:m) = A(k+1:m,k+1:m) - A(k+1:m,k)*A(k,k+1:m)'; */

    }

    /* Perform forward and then back substitution */
    forward_solve( x, A, b);
    back_solve( x, A, x);
}

/*************************************************************
 *
 * FUNCTION: lu_partial_pivot
 *
 * LU factorization with partial pivoting
 * Note that A is overwritten by L and U matrices.
 *************************************************************/
void  lu_partial_pivot( Matrix A, Vector x, const Vector b) {
    int i, j, k, itmp, m=A.m;
    int P[MAXDIM];

    assert( A.m <= MAXDIM);

    /* Initialize k */
    for ( k=0; k<m; ++k) P[k] = k;

    /* First, perform LU factorization. */
    for ( k=0; k<m-1; ++k) {
        /* Find pivot */
        int i_pivot = 0;
        double maxabs=0;

        /* i_pivot = argmax_{i>=k} |A(i,k)| */ 
        for ( i=k; i<m; ++i) {
            double t = fabs( A.vals[i][k]);
            if ( t > maxabs) {
                i_pivot = i;
                maxabs = t;
            }
        }

        if (maxabs==0) {
            fprintf(stderr, "Warning: Matrix is singular.\n");
            continue;
        }

        /* FIXME */
        /* Exchange A(k,:) and A(i_pivot,:); */

        /* Exchange P(k) and P(i_pivot); */
        itmp = P[k];
        P[k] = P[i_pivot];
        P[i_pivot] = itmp;

        /* A(k+1:m,k) = A(k+1:m,k) / A(k,k); */
        for ( i=k+1; i<m; ++i)
            A.vals[i][k] /= A.vals[k][k];

        /* A(k+1:m,k+1:m) = A(k+1:m,k+1:m) - A(k+1:m,k)*A(k,k+1:m)'; */
        for ( i=k+1; i<m; ++i) {
            for ( j=k+1; j<m; ++j)
                A.vals[i][j] -= A.vals[i][k] * A.vals[k][j];
        }
    }

    /* FIXME. Refer to lu_complete_pivot */
    /* Permute b and save output in x. */


    /* Perform forward and then back substitution. */
}


/*************************************************************
 *
 * FUNCTION: lu_complete_pivot
 *
 * LU factorization with complete pivoting
 * Note that A is overwritten to store L and U matrices.
 * y is used for work space.
 *************************************************************/
void  lu_complete_pivot( Matrix A, Vector x, const Vector b, Vector y) {
    int i, j, k, itmp, m=A.m;
    int P[MAXDIM], Q[MAXDIM];

    assert( A.m <= MAXDIM);

    /* Initialize k */
    for ( k=0; k<m; ++k) {
        P[k] = k; Q[k] = k;
    }

    /* First, perform LU factorization. */
    for ( k=0; k<m-1; ++k) {
        /* Find pivot */
        int i_pivot = 0, j_pivot = 0;
        double maxabs=0;

        /* Find pivot */ 
        for ( i=k; i<m; ++i) {
            for ( j=k; j<m; ++j) {
                double t = fabs( A.vals[i][j]);
                if ( t > maxabs) {
                    i_pivot = i;
                    j_pivot = j;
                    maxabs = t;
                }
            }
        }

        /* FIXME. Refer to lu_partial_pivot */
        /* Exchange A(k,:) and A(i_pivot,:); */

        /* Exchange A(:,k) and A(:,j_pivot); */

        /* Exchange P(k) and P(i_pivot); */
        itmp = P[k];
        P[k] = P[i_pivot];
        P[i_pivot] = itmp;

        /* Exchange Q(k) and Q(j_pivot); */
        itmp = Q[k];
        Q[k] = Q[j_pivot];
        Q[j_pivot] = itmp;

        /* A(k+1:m,k) = A(k+1:m,k) / A(k,k); */
        for ( i=k+1; i<m; ++i)
            A.vals[i][k] /= A.vals[k][k];

        /* A(k+1:m,k+1:m) = A(k+1:m,k+1:m) - A(k+1:m,k)*A(k,k+1:m)'; */
        for ( i=k+1; i<m; ++i) {
            for ( j=k+1; j<m; ++j)
                A.vals[i][j] -= A.vals[i][k] * A.vals[k][j];
        }
    }

    /* Permute b and save output in x. */
    permute_vec( x, P, b, 1);

    /* Perform forward and then back substitution. */
    forward_solve( x, A, x);
    back_solve( y, A, x);
    
    permute_vec( x, Q, y, 0);
}

/*************************************************************
 *
 * FUNCTION: forward_solve
 *
 * Perform forward substitution with unit lower triangluar matrix.
 * Note that x can overwrite b.
 *************************************************************/
void  forward_solve( 
    Vector x,
    const Matrix L, 
    const Vector b)
{
    int i, j;
    int n = L.n;

    x.vals[0] =  b.vals[0];
    for (i=1; i<n; i++) {
        double dtemp = 0.0;
        for (j=0; j<i; j++) {
            dtemp = dtemp +  L.vals[i][j] *  x.vals[j];
        }

        x.vals[i] = ( b.vals[i] - dtemp);
    }
    return;
}

/*************************************************************
 *
 * FUNCTION: back_solve
 *
 * Perform back substitution. Note that x can overwrite b.
 *************************************************************/
void  back_solve( 
    Vector x,
    const Matrix U, 
    const Vector b)
{
    int i, j;
    int n = U.n;

    x.vals[n - 1] =  b.vals[n - 1] /  U.vals[n-1][n-1];

    for (i=n-2; i>=0; i--) {
        double dtemp = 0.0;
        for (j=i+1; j<n; j++) {
            dtemp = dtemp +  U.vals[i][j] *  x.vals[j];
        }

        x.vals[i] = ( b.vals[i] - dtemp) /  U.vals[i][i];
    }
    return;
}

/*************************************************************
 *
 * FUNCTION: permute_vec
 *
 * Permute a given vector by computing y = P*x if inv=false or 
 * y = P'*x if inv=true. 
 *************************************************************/

void permute_vec( Vector y, int *P, const Vector x, int inv) {
    int i;

    /* Note that x and y cannot be the same vector. */
    assert( x.vals != y.vals && y.m >= x.m);

    if (!inv) {
        /* Compute y = P'*x */
        for (i=0; i<x.m; ++i)
            y.vals[P[i]] = x.vals[i];
    }
    else {
        /* Compute y = P*x */
        for (i=0; i<x.m; ++i)
            y.vals[i] = x.vals[P[i]];
    }
}

/*************************************************************
 *
 * FUNCTION: rand_matrix_vector
 *
 * Generate a random matrix and vector with entries between -1. and 1.
 *************************************************************/
void rand_matrix_vector( Matrix A, Vector b, int seed) {
    int i,j;
    
    /* Reset seed */
    srand(seed);

    /* Generate random entries for A */
    for (i=0; i<A.m; ++i) {
        b.vals[i] = 0;
        for (j=0; j<A.n; ++j) {
            A.vals[i][j] = 1-2.0*rand()/((double)(RAND_MAX)+1);

            b.vals[i] += (j+1)*A.vals[i][j];
        }
    }
}

/*************************************************************
 *
 * FUNCTION: compute_error
 *
 * Generate a random matrix and vector with entries between -1. and 1.
 *************************************************************/
void compute_error( const Matrix A, const Vector x, const Vector b, 
                    double *err, double *res) {
    int i, j, n=x.m;
    double dtemp[MAXDIM];

    /* Compute error */
    for (i=0; i<n; i++) {
        dtemp[i] =  x.vals[i] - (i+1);
    }
    *err = norm2_col(dtemp, n);

    /* Compute residual. Note that A has to be regenerated. */
    for (i=0; i<n; i++) {
        dtemp[i] = b.vals[i];
        for (j=0; j<n; ++j) {
            dtemp[i] -=  A.vals[i][j] * x.vals[j];
        }
    }
    *res = norm2_col(dtemp, n);
}

/*************************************************************
 *
 * FUNCTION: norm2_col
 *
 * Compute 2-norm of a given vector.
 * Note: It is not a perfect implementation as it does not guard against 
 *       overflow or underflow.
 *************************************************************/

double norm2_col(
    double *vec, 
    int vec_dim)
{
    double v_out;
    int i;

    v_out = 0.0;
    for (i=0; i<vec_dim; i++) {
        v_out = v_out +  vec[i] *  vec[i];
    }
    return sqrt(v_out);
}


/*-----------------------------------------------------------------------------
 * FUNCTION: allocate_matrix - Allocate memory for a given matrix.
 * --------------------------------------------------------------------------*/
/* Disclaimer: The approach used here is not the most efficient way for
 * implementing matrices, but it is adopted here for convenience.
 */
Matrix allocate_matrix(int m, int n) {
    Matrix A;
    double *ptmp;
    int i;

    A.m = m; A.n = n;
    ptmp = (double *)malloc( sizeof(double) * m * n);
    A.vals = (double **)malloc( sizeof(double **) * m);

    for (i=0; i<m; ++i) {
        A.vals[i] = ptmp + i*n;
    }

    return A;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: deallocate_matrix - De-allocate memory for a matrix.
 * --------------------------------------------------------------------------*/
void deallocate_matrix( Matrix A) {
    free( A.vals[0]);
    free( A.vals); A.vals=NULL;
}


/*-----------------------------------------------------------------------------
 * FUNCTION: allocate_vector - Allocate memory for an m vector.
 * --------------------------------------------------------------------------*/
Vector allocate_vector( int m) {
    Vector x;

    x.m = m;
    x.vals = (double *)malloc( sizeof(double) * m);
    return x;
}

/*-----------------------------------------------------------------------------
 * FUNCTION: deallocate_vector - De-allocate the memory for an m vector. 
 * --------------------------------------------------------------------------*/
void deallocate_vector( Vector x) {
    free( x.vals); x.vals=NULL;
}

#include <time.h>
#include <sys/time.h>

/*-----------------------------------------------------------------------------
 * FUNCTION: tictoc - Initialize (1) or get (2) the elapsed time since in seconds
 * --------------------------------------------------------------------------*/
double tictoc (int n)
{
    double    y = 0.0;
    static    struct timeval start_time, end_time; 

    if (n == 1) { 
        gettimeofday(&start_time, NULL);
    }
    else if (n == 2) { 
        gettimeofday(&end_time, NULL);

        y = (end_time.tv_sec - start_time.tv_sec) + (end_time.tv_usec-start_time.tv_usec)*1.e-6;
    }
    return (y); 
}

/*************************************************************
 *
 * FUNCTION: write_vec
 *
 * Write out a vector into file
 *************************************************************/

void  write_vec(FILE *fid, char *name, const double *vec, int n) {
    int i;

    fprintf(fid, "%s=[", name);
    for (i=0; i<n; ++i) {
        fprintf(fid, "%g; ", vec[i]);
    }
    fprintf(fid, "];\n");
}
